from __future__ import annotations

import argparse
import typing

from .. import cli_helpers
from .. import cli_parsing
from . import cd_command

if typing.TYPE_CHECKING:
    from mince import types


def add_dir_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[types.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'dir',
        help='print mince root or dashboard data directory',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=dir_command)
    if include_dashboard_arg:
        parser.add_argument('dashboard', help='dashboard name', nargs='?')
    parser.add_argument(
        '--code', help='print dir of code directory', action='store_true'
    )
    cli_parsing.add_common_args(parser)


def dir_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    if dashboard is None:
        dashboard = args.dashboard
    path = cd_command.get_path(args, dashboard)
    print(path)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from mince.types import StateDatum, UiSpec


def alias_to_state(
    value: str, field: str, ui_spec: UiSpec, *, require: bool = False
) -> StateDatum:
    aliases = ui_spec['inputs'][field]['aliases']
    if value in aliases:
        return aliases[value]
    elif require:
        raise Exception('invalid alias: ' + str(value))
    else:
        return value


def state_to_alias(
    value: StateDatum, field: str, ui_spec: UiSpec, *, require: bool = False
) -> str:
    aliases = ui_spec['inputs'][field]['aliases']
    for alias_value, state_value in aliases.items():
        if value == state_value:
            return alias_value
    if require:
        raise Exception('no alias for state value: ' + str(value))
    else:
        return str(value)



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from mince.figures import ui_figures
from mince import schemas, specs
from . import dashboard_apps
from . import dashboard_shortcuts

if TYPE_CHECKING:
    from typing_extensions import Unpack
    from dash import Dash, html  # type: ignore
    import plotly.graph_objects as go  # type: ignore
    import polars as pl
    import tooltime
    from mince.types import (
        CollectKwargs,
        CollectKwargsPartial,
        CollectionJobSummary,
        UiSpec,
    )


class Dashboard:
    """main object to instantiate or modify when building a dashboard

    the methods exposed in this class are those that may be useful to override
    """

    # class properties

    @classmethod
    def load_spec(cls) -> UiSpec:
        raise NotImplementedError('load_spec() not implemented')

    @classmethod
    async def async_collect_data(
        cls, **kwargs: Unpack[CollectKwargs]
    ) -> dict[str, pl.DataFrame]:
        raise NotImplementedError('async_collect_data() not implemented')

    @classmethod
    def get_default_collect_kwargs(cls) -> CollectKwargsPartial:
        import time

        return {
            'start_time': 0,
            'end_time': int(time.time()),
            'intervals': ['day', 'week', 'month'],
            'skip_incomplete_intervals': False,
            'extra_kwargs': {},
            'extra_kwargs_update': {},
            'verbose': 1,
            'dry': False,
        }

    @classmethod
    def validate(cls, spec: UiSpec, dfs: dict[str, pl.DataFrame]) -> None:
        specs.validate_ui_spec(spec)
        schemas.validate_data_matches_spec(dfs, spec)

    # instance properties

    spec: UiSpec
    dfs: dict[str, pl.DataFrame]
    date_range: tuple[int, int]
    debug: bool
    app: Dash
    t_run_start: float | None
    n_keydowns: int
    job: CollectionJobSummary | None

    def __init__(
        self,
        *,
        dfs: dict[str, pl.DataFrame],
        spec: UiSpec,
        debug: bool = False,
        pdb: bool = False,
        assets_folder: str | None = None,
        date_range: tuple[tooltime.Timestamp, tooltime.Timestamp],
        job: CollectionJobSummary | None = None,
    ):
        import tooltime

        self.validate(dfs=dfs, spec=spec)
        self.spec = spec
        self.spec['default_state']['now'] = tooltime.timestamp_to_date(
            date_range[1]
        )
        self.dfs = dfs
        self.debug = debug
        self.mince_version = specs.get_package_version('mince')
        start_time, end_time = date_range
        self.date_range = (
            tooltime.timestamp_to_seconds(start_time),
            tooltime.timestamp_to_seconds(end_time),
        )
        self.app = dashboard_apps._create_app(
            dashboard=self, assets_folder=assets_folder, pdb=pdb
        )
        self.job = job
        self.n_keydowns = 0
        self.t_run_start = None

    def run(
        self,
        port: str | int | None = None,
        jupyter_mode: str = 'external',
        **kwargs: Any,
    ) -> None:
        import time
        import mince.ops

        if port is None:
            port = str(mince.ops.find_available_port(8052))
        else:
            port = str(port)
        self.t_run_start = time.time()

        try:
            mince.ops.create_pid_file(dashboard=self, port=port)
            self.app.run(
                host='0.0.0.0', jupyter_mode=jupyter_mode, port=port, **kwargs
            )
        finally:
            mince.ops.clear_pid_file(dashboard=self, port=port)

    def get_metadata(self) -> dict[str, Any]:
        import tooltime

        if self.t_run_start is not None:
            t_run_start = tooltime.timestamp_to_iso_pretty(self.t_run_start)
        else:
            t_run_start = None
        return {
            'mince_version': self.mince_version,
            'dashboard_version': self.spec['version'],
            'name': self.spec['name'],
            'description': self.spec['description'],
            'time_started': t_run_start,
            'job': self.job,
        }

    def get_layout(self, inputs: dict[str, html.Div]) -> list[html.Div]:
        from dash import dcc, html

        return [
            dashboard_shortcuts._create_shortcuts_listeners(
                self.spec['shortcuts']
            ),
            dcc.Location(id='url', refresh=False),
            dcc.Store(id='initial-load', data=True),
            dcc.Store(id='radio-group-visibility', data=True),
            html.Div(list(inputs.values()), id='radio-group-row'),
            dcc.Graph(
                id='main-chart',
                config={
                    'responsive': True,
                    'displayModeBar': True,
                    'displaylogo': False,
                    'modeBarButtonsToRemove': [
                        'select2d',
                        'lasso2d',
                        'zoom',
                        'pan',
                        'zoomIn',
                        'zoomOut',
                        'autoScale',
                        'resetScale',
                    ],
                    'modeBarButtonsToAdd': ['toImage'],
                },
            ),
            dashboard_shortcuts._create_help_modal(self.spec['shortcuts']),
            html.Div(id='prevent-focus-trigger'),
        ]

    def get_dataset(self, state: dict[str, Any]) -> str:
        if len(self.dfs) == 1:
            return next(iter(self.dfs.keys()))
        else:
            raise NotImplementedError('get_dataset()')

    def process_shortcuts(
        self,
        display_state: dict[str, Any],
        ui_state: dict[str, Any],
        raw_shortcut: dict[str, Any],
    ) -> None:
        dashboard_shortcuts._process_keyboard_shortcuts(
            display_state=display_state,
            ui_state=ui_state,
            raw_shortcut=raw_shortcut,
            data_date_range=self.date_range,
            ui_spec=self.spec,
        )

    def create_chart(self, state: dict[str, Any]) -> go.Figure:
        data_name = self.get_dataset(state)
        df = self.dfs[data_name]

        title = self.create_title(
            df=df,
            state=state,
            data_date_range=self.date_range,
            ui_spec=self.spec,
        )

        if state['format'] in ['line', 'line %', 'area', 'area %']:
            return ui_figures.create_time_series_fig(
                df, state, self.debug, ui_spec=self.spec, title=title
            )
        elif state['format'] == 'tree':
            return ui_figures.create_treemap_fig(
                df,
                state,
                ui_spec=self.spec,
                data_date_range=self.date_range,
                title=title,
            )
        else:
            raise Exception('invalid format: ' + str(state['format']))

    def create_title(
        self,
        df: pl.DataFrame,
        state: dict[str, Any],
        data_date_range: tuple[int, int],
        ui_spec: UiSpec,
    ) -> str | None:
        return None



from __future__ import annotations

import typing

import toolcache

from mince import dashboards
from mince import schemas

if typing.TYPE_CHECKING:
    import polars as pl
    from mince import types


@toolcache.cache(cachetype='memory', hash_include_args=['state'])
def _prepare_time_series_data(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    ui_spec: types.UiSpec,
) -> pl.DataFrame:
    import polars as pl
    import tooltime

    # filter end of timerange
    now = tooltime.timestamp_to_seconds(state['now'])
    df = df.filter(pl.col.timestamp.cast(pl.Int64) <= now * 1_000)

    # filter by metric
    df_metric = dashboards.get_df_metric(state, ui_spec)
    df = df.filter(pl.col.metric == df_metric)

    # aggregate by irrelevant dimensions
    df = _aggregate_by_irrelevant_dimensions(
        df,
        state,
        ui_spec['schema']['columns'],
        ui_spec,
    )

    # aggregate by time interval
    df = _aggregate_by_interval(df, state)

    # add age column
    if state['xalign'] == 'age':
        df = _add_age(df, state)

    # get cumulative value from t=0
    if (
        state['cumulative'] == 'cumulative (t=0)'
        and ui_spec['metrics'][df_metric]['type'] == 'sum'
    ):
        df = df.sort('timestamp').with_columns(
            value=pl.cum_sum('value').over(
                [
                    col
                    for col in schemas.get_series_columns(
                        ui_spec['schema']['columns']
                    )
                    if col != 'value' and col in df
                ]
            )
        )

    # filter by date range
    df = _filter_by_time_range(df, state)

    # get cumulative value from inside window
    if (
        state['cumulative'] == 'cumulative (window)'
        and ui_spec['metrics'][df_metric]['type'] == 'sum'
    ):
        df = df.sort('timestamp').with_columns(
            value=pl.cum_sum('value').over(
                [
                    col
                    for col in schemas.get_series_columns(
                        ui_spec['schema']['columns']
                    )
                    if col != 'value' and col in df
                ]
            )
        )

    # format-specific transformations
    if state['format'] == 'line':
        if state['total'] == 'total':
            df = _add_total_entry(df, state)
        if state['ynormalize'] == 'relative':
            df = _convert_to_relative(df, state)
    elif state['format'] == 'line %':
        df = _convert_to_percent(df, state)
    elif state['format'] == 'area':
        pass
    elif state['format'] == 'area %':
        df = _convert_to_percent(df, state)
    else:
        raise Exception('invalid format')

    # sort data
    df = _sort_data(df, state)

    return df


def _aggregate_by_irrelevant_dimensions(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    schema: dict[str, types.ColumnType],
    ui_spec: types.UiSpec,
) -> pl.DataFrame:
    import polars as pl

    agg_columns = [
        column for column in ui_spec['groupings'] if column != state['grouping']
    ]

    return (
        df.group_by(
            [col for col in df.columns if col not in agg_columns + ['value']],
            maintain_order=True,
        )
        .agg(pl.sum('value'))
        .with_columns(
            *[pl.lit('all').alias(agg_column) for agg_column in agg_columns]
        )
        .select(list(schema.keys()))
    )


def _aggregate_by_interval(
    df: pl.DataFrame, state: dict[str, typing.Any]
) -> pl.DataFrame:
    import polars as pl

    if len(df) == 0:
        raise Exception('empty dataframe')

    available_intervals = list(df['interval'].unique())

    if state['sample_interval'] == 'date':
        if 'day' in available_intervals:
            return df.filter(pl.col.interval == 'day')
        elif 'point' in available_intervals:
            return df.filter(pl.col.interval == 'point')
        else:
            raise Exception('date-level data not available')
    elif state['sample_interval'] in ['week', 'month']:
        if state['sample_interval'] in available_intervals:
            return df.filter(pl.col.interval == state['sample_interval'])
        else:
            # add sample_interval column
            if state['sample_interval'] == 'week':
                unix_seconds = pl.col.timestamp.cast(pl.Int64) / 1_000
                unix_week = ((unix_seconds + 4 * 86400) / 86400 / 7).floor()
                week = (unix_week * 86400 * 7 * 1_000 - 4 * 86400 * 1_000).cast(
                    pl.Datetime('ms')
                )
                df = df.with_columns(week=week)
            elif state['sample_interval'] == 'month':
                df = df.with_columns(month=pl.col.timestamp.dt.month_start())
            else:
                raise Exception('invalid sample interval')

            # perform aggregation
            group_columns = [
                state['grouping'],
                state['sample_interval'],
                'interval',
            ]
            if 'day' in available_intervals:
                return (
                    df.filter(pl.col.interval == 'day')
                    .group_by(group_columns, maintain_order=True)
                    .agg(pl.sum('value'))
                    .with_columns(
                        pl.col(state['sample_interval']).alias('timestamp'),
                        interval=pl.lit(state['sample_interval']),
                    )
                )
            elif 'point' in available_intervals:
                return (
                    df.filter(pl.col.interval == 'point')
                    .group_by(group_columns, maintain_order=True)
                    .agg(pl.first('value'))
                    .with_columns(
                        pl.col(state['sample_interval']).alias('timestamp'),
                    )
                )
            else:
                raise Exception('date-level data not available')
    else:
        raise Exception('invalid interval: ' + str(state['sample_interval']))


def _add_age(df: pl.DataFrame, state: dict[str, typing.Any]) -> pl.DataFrame:
    import polars as pl

    x = 'timestamp'
    raw_new_time_index = df[[x]].unique(x).sort(x).with_row_index()
    new_time_index = dict(raw_new_time_index[[x, 'index']].rows())

    group_starts = (
        df.group_by(state['grouping'], x, maintain_order=True)
        .agg(value=pl.sum('value'))
        .filter(pl.col.value > 0)
        .group_by(state['grouping'], maintain_order=True)
        .agg(
            new_x_start=pl.min(x).replace(new_time_index, return_dtype=pl.Int64)
        )
    )

    output = (
        df.join(group_starts, on=state['grouping'])
        .with_columns(
            age=pl.col(x).replace(new_time_index, return_dtype=pl.Int64)
            - pl.col.new_x_start
        )
        .filter(pl.col.age >= 0)
    ).drop('new_x_start')

    return output


def _filter_by_time_range(
    df: pl.DataFrame, state: dict[str, typing.Any]
) -> pl.DataFrame:
    import polars as pl
    import tooltime

    if state['time_window'] != 'all':
        if state['xalign'] == 'age':
            x = state['sample_interval']
            if x == 'date':
                age_unit = 86400
            elif x == 'week':
                age_unit = 7 * 86400
            elif x == 'month':
                age_unit = 30 * 86400
            else:
                raise Exception(
                    'invalid sample_interval: ' + str(state['sample_interval'])
                )
            max_age = (
                tooltime.timelength_to_seconds(state['time_window']) / age_unit
            )
            max_age = max(max_age, 1)
            df = df.filter(pl.col.age <= max_age)
        else:
            start_time = tooltime.timestamp_to_seconds(
                state['now']
            ) - tooltime.timelength_to_seconds(state['time_window'])
            df = df.filter((pl.col.timestamp >= start_time * 1e3))
    return df


def _add_total_entry(
    df: pl.DataFrame, state: dict[str, typing.Any]
) -> pl.DataFrame:
    import polars as pl

    if state['format'] == 'line' and state['xalign'] == 'age':
        x_dimension = 'age'
    else:
        x_dimension = 'timestamp'
    new_columns: dict[str, str | pl.Expr] = {
        state['grouping']: pl.lit('TOTAL').alias(state['grouping']),
        'interval': 'interval',
        'timestamp': 'timestamp',
        x_dimension: x_dimension,
        'value': 'value',
    }
    unused_columns = [
        column for column in df.columns if column not in new_columns
    ]
    if len(unused_columns) > 0:
        df = df.drop(unused_columns)
    df_total = (
        df.group_by(x_dimension, maintain_order=True)
        .agg(
            pl.sum('value'),
            pl.first('interval'),
            # pl.first('timestamp'),
        )
        .select(*[new_columns[key] for key in df.columns])
    )
    df = pl.concat([df, df_total])

    return df


def _convert_to_relative(
    df: pl.DataFrame, state: dict[str, typing.Any]
) -> pl.DataFrame:
    import polars as pl

    initial_value_per_group = df.group_by(
        state['grouping'], maintain_order=True
    ).agg(initial_value=pl.first('value'))
    df = df.filter(pl.col.value.is_not_null() & (pl.col.value > 0))
    df = df.join(initial_value_per_group, on=state['grouping'], how='left')
    df = df.filter(pl.col.initial_value >= 0.00001)
    df = (
        df.with_columns(value=pl.col.value / pl.col.initial_value)
        .with_columns(value=pl.col.value.replace({float('inf'): 0}))
        .fill_nan(None)
        .fill_null(0)
        .drop('initial_value')
    )
    return df


def _convert_to_percent(
    df: pl.DataFrame, state: dict[str, typing.Any]
) -> pl.DataFrame:
    import polars as pl

    if state['format'] == 'line' and state['xalign'] == 'age':
        x_dimension = 'age'
    else:
        x_dimension = 'timestamp'
    df = df.with_columns(value=pl.col.value / pl.sum('value').over(x_dimension))
    return df


def _sort_data(df: pl.DataFrame, state: dict[str, typing.Any]) -> pl.DataFrame:
    import polars as pl

    ranks = (
        df.group_by(state['grouping'])
        .agg(pl.sum('value'))
        .sort('value', descending=True)
        .with_row_index()
        .rename({'index': 'rank'})
    )
    return (
        df.join(
            ranks[[state['grouping'], 'rank']],
            on=state['grouping'],
            how='left',
        )
        .sort('timestamp', 'rank')
        .drop('rank')
    )



from __future__ import annotations

import typing

import mince

if typing.TYPE_CHECKING:
    import polars as pl


def compute_time_chunks(
    start_time: float,
    end_time: float,
    chunk_size: str,
    interval: mince.Interval,
    tags: dict[str, str],
    get_chunk_time_range: typing.Callable[..., tuple[int, int]] | None,
    skip_incomplete_intervals: bool,
) -> pl.DataFrame:
    import tooltime

    # convert chunk_size to be integer multiple of interval
    interval_chunk_size = round_up_chunk_size(chunk_size, interval)

    # adjust target data range
    if get_chunk_time_range is None:
        use_start, use_end = start_time, end_time
    else:
        use_start, use_end = get_chunk_time_range(
            start_time=start_time,
            end_time=end_time,
            interval=interval,
            tags=tags,
        )

    return tooltime.get_intervals(
        start=use_start,
        end=use_end,
        interval=interval_chunk_size,
        include_incomplete=not skip_incomplete_intervals,
        clip_inward=True,
    )


def round_up_chunk_size(chunk_size: str, interval: mince.Interval) -> str:
    """round up chunk size so that it can be evenly divided by interval
    - each chunk must be an integer multiple of the sampling interval
    - chunks in a chunkset do not need to all be the same size
    """
    import math

    chunk_count = int(chunk_size[:-1])
    chunk_unit = chunk_size[-1]

    if interval == 'hour':  # type: ignore
        # all chunk units are evently divisible into hours
        return chunk_size
    elif interval == 'day':
        if chunk_unit == 'h':
            return str(math.ceil(chunk_count / 24)) + 'd'
        elif chunk_unit == 'd':
            return chunk_size
        elif chunk_unit == 'w':
            return chunk_size
        elif chunk_unit == 'M':
            return chunk_size
        elif chunk_unit == 'y':
            return chunk_size
        else:
            raise Exception('invalid chunk unit')
    elif interval == 'week':
        if chunk_unit == 'h':
            return str(math.ceil(chunk_count / 7 / 24)) + 'w'
        elif chunk_unit == 'd':
            return str(math.ceil(chunk_count / 7)) + 'w'
        elif chunk_unit == 'w':
            return chunk_size
        elif chunk_unit == 'M':
            return str(math.ceil(4.43 * chunk_count)) + 'w'
        elif chunk_unit == 'y':
            return str(math.ceil(52.3 * chunk_count)) + 'w'
        else:
            raise Exception('invalid chunk unit')
    elif interval == 'month':
        if chunk_unit == 'h':
            return str(math.ceil(chunk_count / 28 / 24)) + 'M'
        elif chunk_unit == 'd':
            return str(math.ceil(chunk_count / 28)) + 'M'
        elif chunk_unit == 'w':
            return str(math.ceil(chunk_count / 4)) + 'M'
        elif chunk_unit == 'M':
            return chunk_size
        elif chunk_unit == 'y':
            return chunk_size
        else:
            raise Exception('invalid chunk unit')
    elif interval == 'year':
        if chunk_unit == 'h':
            return str(math.ceil(chunk_count / 365 / 24)) + 'y'
        elif chunk_unit == 'd':
            return str(math.ceil(chunk_count / 365)) + 'y'
        elif chunk_unit == 'w':
            return str(math.ceil(chunk_count / 52)) + 'y'
        elif chunk_unit == 'M':
            return str(math.ceil(chunk_count / 12)) + 'y'
        elif chunk_unit == 'y':
            return chunk_size
        else:
            raise Exception('invalid chunk unit')
    else:
        raise Exception('invalid sampling interval: ' + str(interval))



from __future__ import annotations

import argparse
import typing

from ... import storage
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_data_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[types.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'data',
        help='view dashboard data',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=data_command)
    if include_dashboard_arg:
        parser.add_argument(
            'dashboard',
            help='dashboard to get data of',
        )
    parser.add_argument(
        '-i',
        '--interactive',
        help='load data in interactive python session',
        action='store_true',
    )
    cli_parsing.add_common_args(parser)


def data_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    if dashboard is None:
        dashboard = args.dashboard
    data_dir = storage.resolve_data_dir(
        dashboard=dashboard,
        registry={'path': args.registry, 'validate': not args.no_validate},
    )
    job, dfs = storage.load_dashboard_data(data_dir=data_dir)

    if args.interactive:
        cli_helpers.open_interactive_session(variables=dfs)
    else:
        first = True
        for key, df in dfs.items():
            print()
            if first:
                first = False
            else:
                print()
            print('name =', key)
            print(df)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl
    import tooltime


def get_date_range(df: pl.DataFrame) -> tuple[int, int]:
    import datetime
    import tooltime

    if len(df) == 0:
        raise Exception('empty dataframe')

    start_time: datetime.datetime = df['timestamp'].min()  # type: ignore
    end_time: datetime.datetime = df['timestamp'].max()  # type: ignore

    return (
        tooltime.timestamp_to_seconds(start_time),
        tooltime.timestamp_to_seconds(end_time),
    )


def week_column(*, time_column: str | None = None) -> pl.Expr:
    if time_column is None:
        time_column = 'timestamp'
    return (
        pl.col(time_column)
        .dt.offset_by('1d')
        .dt.truncate('1w')
        .dt.offset_by('-1d')
        .alias('week')
    )


def end_timestamp_column(*, time_column: str | None = None) -> pl.Expr:
    if time_column is None:
        time_column = 'timestamp'
    return (
        pl.when(interval='year')
        .then(pl.col(time_column).dt.truncate('1y').dt.offset_by('1y'))
        .when(interval='month')
        .then(pl.col(time_column).dt.truncate('1mo').dt.offset_by('1mo'))
        .when(interval='week')
        .then(week_column().dt.offset_by('1d'))
        .when(interval='day')
        .then(pl.col(time_column).dt.truncate('1d').dt.offset_by('1d'))
        .when(interval='hour')
        .then(pl.col(time_column).dt.truncate('1d').dt.offset_by('1d'))
        .otherwise(None)
        .alias('end_timestamp')
    )


def interval_completed_column(
    cutoff: tooltime.Timestamp,
    *,
    time_column: str | None = None,
) -> pl.Expr:
    import tooltime

    dt_cutoff = tooltime.timestamp_to_datetime(cutoff)
    return dt_cutoff < end_timestamp_column(time_column=time_column)



from __future__ import annotations

import argparse
import typing

import mince
from ... import storage
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_register_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'register',
        help='register dashboard',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=register_command)
    parser.add_argument(
        'dashboard_class',
        metavar='MODULE:CLASS',
        help='Dashboard class',
    )
    cli_parsing.add_common_args(parser)


def register_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    registry: types.RegistryFile = {
        'path': args.registry,
        'validate': not args.no_validate,
    }
    if dashboard is None:
        dashboard = args.dashboard_class
    storage.register_dashboard(
        dashboard=dashboard,
        registry=registry,
    )



from .dashboard_apps import *
from .dashboard_class import *
from .dashboard_classes import *
from .dashboard_inputs import *
from .dashboard_shortcuts import *
from .dashboard_state import *
from .dashboard_urls import *



"""
path layout is identical on local filesystem vs http server

Layout:
- {root_dir}/{dashboard}/
    - fragments/
    - latest.json --> symlink to a data.json
    - jobs
        - 2024-01-05/
            - data.json
            - datum1.parquet
            - datum2.parquet
            - datum3.parquet
        - 2024-01-04/
            - data.json
            - datum1.parquet
            - datum2.parquet
            - datum3.parquet
        - 2024-01-03/
            - data.json
            - datum1.parquet
            - datum2.parquet
            - datum3.parquet
"""

from __future__ import annotations

import os
import typing

from . import registry as registry_module

if typing.TYPE_CHECKING:
    from mince import types


def get_mince_root() -> str:
    path = os.environ.get('MINCE_ROOT')
    if path is None:
        print('MINCE_ROOT not set, using ~/mince as default mince root')
        path = '~/data/mince'
    return os.path.expanduser(path)


def get_registry_path() -> str:
    path = os.environ.get('MINCE_REGISTRY_PATH')
    if path is None:
        path = os.path.join(get_mince_root(), 'mince_registry.json')
    return os.path.expanduser(path)


#
# # registry resolution
#


def resolve_data_dir(
    dashboard: str | typing.Type[types.Dashboard],
    registry: types.RegistryReference = None,
) -> str:
    """input is either a registry entry name or a Dashboard subclass"""
    import mince

    if isinstance(dashboard, str):
        registry_entry = registry_module.get_registry_entry(
            dashboard, registry=registry
        )
        registry_data_dir = registry_entry['collect_kwargs'].get('data_dir')
        if registry_data_dir is not None:
            return registry_data_dir
        else:
            return get_dashboard_data_dir(dashboard)
    elif type(dashboard) is type and issubclass(dashboard, mince.Dashboard):
        spec = dashboard.load_spec()
        return get_dashboard_data_dir(spec['name'])
    else:
        raise Exception('invalid type')


#
# # absolute paths
#


def get_dashboard_data_dir(dashboard: str | None) -> str:
    if dashboard is None:
        raise Exception('specify data_dir or dashboard')
    return os.path.join(get_mince_root(), dashboard)


def get_job_dir(
    *,
    timestamp: int | float,
    dashboard: str | None = None,
    data_dir: str | None = None,
) -> str:
    if data_dir is None:
        data_dir = get_dashboard_data_dir(dashboard=dashboard)
    return os.path.join(data_dir, get_job_reldir(timestamp=timestamp))


def get_job_metadata_path(
    *,
    timestamp: int | float,
    dashboard: str | None = None,
    data_dir: str | None = None,
) -> str:
    if data_dir is None:
        data_dir = get_dashboard_data_dir(dashboard=dashboard)
    return os.path.join(data_dir, get_job_metadata_relpath(timestamp=timestamp))


def get_job_datum_path(
    *,
    datum_name: str,
    timestamp: int | float,
    dashboard: str | None = None,
    data_dir: str | None = None,
) -> str:
    if data_dir is None:
        data_dir = get_dashboard_data_dir(dashboard=dashboard)
    relpath = get_job_datum_relpath(datum_name=datum_name, timestamp=timestamp)
    return os.path.join(data_dir, relpath)


def get_job_latest_path(
    *,
    dashboard: str | None = None,
    data_dir: str | None = None,
) -> str:
    if data_dir is None:
        data_dir = get_dashboard_data_dir(dashboard=dashboard)
    return os.path.join(data_dir, get_latest_job_relpath())


#
# # relative paths
#


def get_job_reldir(timestamp: int | float) -> str:
    import datetime

    dt = datetime.datetime.fromtimestamp(timestamp)
    dt_str = dt.strftime('%Y-%m-%d_%H-%M-%S.%f')
    return os.path.join('jobs', dt_str)


def get_job_metadata_relpath(timestamp: int | float) -> str:
    job_dir = get_job_reldir(timestamp=timestamp)
    return os.path.join(job_dir, 'data.json')


def get_job_datum_relpath(datum_name: str, timestamp: int | float) -> str:
    job_dir = get_job_reldir(timestamp=timestamp)
    return os.path.join(job_dir, datum_name + '.parquet')


def get_latest_job_relpath() -> str:
    return 'latest.json'



from __future__ import annotations

import os
import typing

if typing.TYPE_CHECKING:
    from mince import Dashboard


pid_dir = '/tmp/mince/running'


def get_pid_path(*, dashboard: Dashboard, port: int | str, pid: int) -> str:
    pid_template = os.path.join(pid_dir, '{dashboard}__port_{port}__pid_{pid}')
    return pid_template.format(
        dashboard=dashboard.spec['name'],
        port=port,
        pid=pid,
    )


def create_pid_file(dashboard: Dashboard, port: str) -> None:
    pid = os.getpid()
    path = get_pid_path(dashboard=dashboard, port=port, pid=pid)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as _:
        pass


def clear_pid_file(dashboard: Dashboard, port: str) -> None:
    pid = os.getpid()
    path = get_pid_path(dashboard=dashboard, port=port, pid=pid)
    try:
        os.remove(path)
    except OSError:
        pass


def get_running_servers() -> list[dict[str, typing.Any]]:
    servers = []
    for filename in os.listdir(pid_dir):
        # parse pid filename
        pieces = filename.split('__')
        server: dict[str, typing.Any] = {
            'name': pieces[0],
            'port': pieces[1].split('_')[1],
            'pid': pieces[2].split('_')[1],
        }

        # get server metadata
        try:
            server['metadata'] = get_server_metadata(server)
        except Exception:
            os.remove(os.path.join(pid_dir, filename))
            continue

        servers.append(server)

    return servers


def get_server_metadata(server: dict[str, typing.Any]) -> dict[str, typing.Any]:
    import requests

    url = 'http://localhost:' + str(server['port']) + '/metadata'
    response = requests.get(url)
    return response.json()  # type: ignore


def kill_server(
    *,
    name: str | None = None,
    pid: str | int | None = None,
    port: str | int | None = None,
) -> None:
    import signal

    if pid is not None:
        pid = str(pid)
    if port is not None:
        port = str(port)

    if name is None and pid is None and port is None:
        raise Exception('must specify name or pid or port')

    candidates = []
    for server in get_running_servers():
        if name is not None and name != server['name']:
            continue
        if port is not None and port != server['port']:
            continue
        if pid is not None and pid != server['pid']:
            continue
        candidates.append(server)

    if len(candidates) == 0:
        print('[no matching dashboards]')
    elif len(candidates) == 1:
        print('killing 1 dashboard')
    else:
        print('killing', len(candidates), 'dashboards')

    for server in candidates:
        os.kill(int(server['pid']), signal.SIGTERM)



from __future__ import annotations

import typing
import mince

if typing.TYPE_CHECKING:
    import polars as pl


class PiecewiseCollector:
    path_template: str
    tag_names: list[str]

    @classmethod
    def get_chunk_time_range(
        cls,
        start_time: int,
        end_time: int,
        interval: mince.Interval,
        tags: dict[str, str],
    ) -> tuple[int, int]:
        return start_time, end_time

    @classmethod
    async def async_collect_df(
        cls,
        start_time: int,
        end_time: int,
        interval: mince.Interval,
        tags: dict[str, str],
    ) -> pl.DataFrame:
        raise NotImplementedError()

    @classmethod
    async def async_collect_data(
        cls,
        data_dir: str,
        start_time: float,
        end_time: float,
        intervals: list[mince.Interval],
        skip_incomplete_intervals: bool,
        extra_kwargs: dict[str, typing.Any],
        verbose: int,
        dry: bool,
    ) -> dict[str, pl.DataFrame]:
        return await mince.ops.async_collect_data_piecewise(
            data_dir=data_dir,
            start_time=start_time,
            end_time=end_time,
            intervals=intervals,
            skip_incomplete_intervals=skip_incomplete_intervals,
            chunk_size=extra_kwargs['chunk_size'],
            skip_existing_files=extra_kwargs['skip_existing_files'],
            max_concurrent_queries=extra_kwargs['max_concurrent_queries'],
            async_collect_df=cls.async_collect_df,
            get_chunk_time_range=cls.get_chunk_time_range,
            path_template=cls.path_template,
            tags={
                tag_name: extra_kwargs[tag_name] for tag_name in cls.tag_names
            },
            verbose=verbose,
            dry=dry,
        )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from mince import types


def parse_url(
    url_search: str,
    display: dict[str, str],
    defaults: dict[str, types.StateDatum | None],
) -> None:
    import urllib.parse

    parsed = urllib.parse.parse_qs(urllib.parse.urlparse(url_search).query)
    for key in defaults.keys():
        if key in parsed:
            display[key] = parsed[key][0]
    if 'now' in parsed:
        display['now'] = parsed['now'][0]


def create_new_url(
    display: dict[str, str],
    defaults: dict[str, types.StateDatum | None],
) -> str:
    params: list[str] = []
    for key, value in defaults.items():
        if display[key] != str(value):
            params.append(key + '=' + display[key])

    return '?' + '&'.join(params) if params else ''
